#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "kanji.h"

struct kanji_bitmap {
  struct kanji_bitmap *next;
  unsigned long long code;
  int width;
  int height;
  unsigned char *bits;
};

static struct kanji_bitmap *kanji_list = NULL;

static struct kanji_bitmap *add_kanji(unsigned long long code,
				      int width, int height,
				      unsigned char *bits)
{
  struct kanji_bitmap *k;
  k = malloc(sizeof(*k));
  memset(k, 0, sizeof(*k));
  k->code = code;
  k->width = width;
  k->height = height;
  k->bits = bits;
  k->next = kanji_list;
  kanji_list = k;
  return k;
}

static struct kanji_bitmap *search_kanji(unsigned long long code)
{
  struct kanji_bitmap *k, **kp;
  for (kp = &kanji_list; *kp; kp = &((*kp)->next)) {
    k = *kp;
    if (k->code == code) {
      *kp = k->next;
      k->next = kanji_list;
      kanji_list = k;
      return k;
    }
  }
  return NULL;
}

static char *skip_space(char *p)
{
  while (*p) {
    if (!isspace(*p))
      break;
    p++;
  }
  return p;
}

static char *skip_nospace(char *p)
{
  while (*p) {
    if (isspace(*p))
      break;
    p++;
  }
  return p;
}

int kanji_init(char *filename)
{
  FILE *fp;
  char buf[128], *p;
  unsigned long long code;
  int width, height, size, x, y;
  unsigned char *bits;

  fp = fopen(filename, "r");
  if (fp == NULL)
    return -1;

  code = width = height = y = 0;
  bits = NULL;

  while (fgets(buf, sizeof(buf), fp) != NULL) {
    if (!strncmp(buf, "BEGIN", 5)) {
      p = buf;
      p = skip_nospace(p); p = skip_space(p); code   = strtoll(p, NULL, 16);
      p = skip_nospace(p); p = skip_space(p); width  = atoi(p);
      p = skip_nospace(p); p = skip_space(p); height = atoi(p);
      size = ((width + 7) / 8) * height;
      bits = malloc(size);
      memset(bits, 0, size);
      y = 0;
      continue;
    }
    if (!strncmp(buf, "END", 3)) {
      if (bits) {
	add_kanji(code, width, height, bits);
	bits = NULL;
      }
    }
    if (bits && isxdigit(buf[0])) {
      p = skip_nospace(buf); p = skip_space(p);
      for (x = 0; x < width; x++) {
	if (*p == '\0') break;
	if (*p == '*') {
	  bits[y * ((width + 7) / 8) + (x / 8)] |= (1 << (7 - (x % 8)));
	}
	p++;
      }
      y++;
      continue;
    }
  }
  return 0;
}

int kanji_get_size(unsigned char top)
{
  if ((top & 0x80) == 0x00) return 1;
  if ((top & 0xe0) == 0xc0) return 2;
  if ((top & 0xf0) == 0xe0) return 3;
  if ((top & 0xf8) == 0xf0) return 4;
  if ((top & 0xfc) == 0xf8) return 5;
  if ((top & 0xfe) == 0xfc) return 6;
  return 0;
}

int kanji_get_width(unsigned long long code)
{
  struct kanji_bitmap *k;
  k = search_kanji(code);
  return k ? k->width : 0;
}

int kanji_get_height(unsigned long long code)
{
  struct kanji_bitmap *k;
  k = search_kanji(code);
  return k ? k->height : 0;
}

int kanji_get_bit(unsigned long long code, int x, int y)
{
  struct kanji_bitmap *k;

  k = search_kanji(code);
  if (k == NULL) return 0;
  if ((x < 0) || (x >= k->width) || (y < 0) || (y >= k->height))
    return 0;

  return (k->bits[y * ((k->width + 7) / 8) + (x / 8)] & (1 << (7 - (x % 8)))) ? 1 : 0;
}
