#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <GL/glut.h>

#include "argument.h"
#include "socket.h"
#include "vector.h"
#include "pattern.h"
#include "kanji.h"
#include "Obj.h"
#include "Saucer.h"
#include "Block.h"
#include "Log.h"
#include "Sock.h"

static int saucer_num = 3;
static int track_num = 1;
static int sector_num = 3;
static int auto_rate = 0;
static int multi_sock = 0;
static double pattern_rate = 1.0;

struct {
  double angle_dir;
  double angle_up;
  double distance;
  double angle_dir_v;
  vector_t position;
  vector_t direction;
  vector_t updir;
} camera;

static int gl_camera()
{
  vector_t p;
  double rad;

  vector_set(&camera.position, camera.distance, 0.0, 0.0);

  rad = ((2.0 * M_PI) / 360.0) * -camera.angle_up;
  p = camera.position;
  camera.position.z = cos(rad) * p.z - sin(rad) * p.x;
  camera.position.x = sin(rad) * p.z + cos(rad) * p.x;

  rad = ((2.0 * M_PI) / 360.0) * camera.angle_dir;
  p = camera.position;
  camera.position.x = cos(rad) * p.x - sin(rad) * p.y;
  camera.position.y = sin(rad) * p.x + cos(rad) * p.y;

  vector_mul(&camera.direction, &camera.position, -1.0);
  vector_set(&camera.updir, 0.0, 0.0, 1.0);

  return 0;
}

static int gl_init()
{
  GLfloat ambient[]  = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat diffuse[]  = { 0.7, 0.7, 0.7, 1.0 };
  GLfloat specular[] = { 0.1, 0.1, 0.1, 1.0 };

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glEnable(GL_DEPTH_TEST);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glEnable(GL_LIGHTING);

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

  glLightfv(GL_LIGHT0, GL_AMBIENT,  ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  return 0;
}

static int gl_light()
{
  GLfloat position[] = { 0.0, 0.0, 2.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  glEnable(GL_LIGHT0);
  return 0;
}

static ObjList saucers;
static ObjList users;
static ObjList genres;

static int set_saucers(int n)
{
  Saucer saucer;
  ObjListData obj;
  static int number = 0;
  int i, len;
  vector_t *points;
  vector_t *angles;

  number += n;

  len = ObjList_GetLength(saucers);
  points = malloc(sizeof(*points) * len);
  angles = malloc(sizeof(*angles) * len);

  i = 0;
  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);
    vector_copy(&points[i], Saucer_GetPosition(saucer));
    vector_copy(&angles[i], Saucer_GetAngle(saucer));
    i++;
  }

  pattern_set(number, len, points, angles);

  for (i = 0; i < len; i++) {
    vector_mul(&points[i], &points[i], pattern_rate);
  }

  i = 0;
  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);
    Saucer_Move(saucer, 10, points[i].x, points[i].y, points[i].z);
    Saucer_Rotate(saucer, 10, angles[i].x, angles[i].y, angles[i].z);
    i++;
  }

  free(points);
  free(angles);

  return 0;
}

static Saucer make_saucer(char *name)
{
  Saucer saucer;
  Block block;
  Genre genre;
  Question question;
  ObjList questions;
  ObjListData obj, obj2;
  int t, s;

  saucer = Saucer_Create(0.0, 0.0, 0.0, track_num, sector_num, name);

  s = 0;
  for (obj = ObjList_GetStart(genres);
       !ObjList_IsEndEdge(genres, obj);
       obj = ObjListData_GetNext(obj)) {
    genre = (Genre)ObjListData_GetObj(obj);
    questions = Genre_GetQuestions(genre);
    t = 0;
    for (obj2 = ObjList_GetStart(questions);
	 !ObjList_IsEndEdge(questions, obj2);
	 obj2 = ObjListData_GetNext(obj2)) {
      question = (Question)ObjListData_GetObj(obj2);
      block = Saucer_GetBlock(saucer, t, s);
      if (block)
	Block_SetQuestion(block, question);
      t++;
    }
    s++;
  }

  ObjList_InsertObjToEnd(saucers, saucer, NULL);

  return saucer;
}

static int make_saucers()
{
  int i;

  saucers = ObjList_Create();
  users = ObjList_Create();
  genres = ObjList_Create();

  for (i = 0; i < saucer_num; i++)
    make_saucer("noname");

  set_saucers(0);

  return 0;
}

static int init()
{
  srand(time(NULL));

  camera.angle_dir   = 0.0;
  camera.angle_up    = 20.0;
  camera.distance    = 5.0;
  camera.angle_dir_v = 0.2;

  make_saucers();

  gl_camera();
  gl_init();

  return 0;
}

static void gl_display()
{
  Saucer saucer;
  ObjListData obj;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();

  gl_camera();
  gluLookAt(camera.position.x, camera.position.y, camera.position.z,
	    camera.direction.x, camera.direction.y, camera.direction.z,
	    camera.updir.x, camera.updir.y, camera.updir.z);
  gl_light();

  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);

    glPushMatrix();
    Saucer_MakePolygon(saucer, genres);
    glPopMatrix();
  }

  glutSwapBuffers();
}

static void gl_keyboard(unsigned char key, int x, int y)
{
  Saucer saucer;
  Block block;
  ObjListData obj;
  int n, t, s;

  switch (key) {
  case '\033':
  case 'q':
    exit(0);
  case 'i': camera.angle_up  += 0.5;  break;
  case 'k': camera.angle_up  -= 0.5;  break;
  case 'j': camera.angle_dir -= 1.0;  break;
  case 'l': camera.angle_dir += 1.0;  break;
  case 'o': camera.distance  += 0.03; break;
  case 'u': camera.distance  -= 0.03; break;

  case 'd': camera.angle_dir_v += 0.02; break;
  case 'f': camera.angle_dir_v -= 0.02; break;

  case 'z': if (auto_rate >    0) auto_rate--; break;
  case 'x': if (auto_rate < 1000) auto_rate++; break;

  case '1': pattern_rate -= 0.04; set_saucers(0); break;
  case '2': pattern_rate += 0.04; set_saucers(0); break;

  case 't':
    for (obj = ObjList_GetStart(saucers);
	 !ObjList_IsEndEdge(saucers, obj);
	 obj = ObjListData_GetNext(obj)) {
      saucer = (Saucer)ObjListData_GetObj(obj);
      Saucer_AddTrackNum(saucer, 1);
    }
    track_num++;
    break;
  case 'g':
    for (obj = ObjList_GetStart(saucers);
	 !ObjList_IsEndEdge(saucers, obj);
	 obj = ObjListData_GetNext(obj)) {
      saucer = (Saucer)ObjListData_GetObj(obj);
      Saucer_AddSectorNum(saucer, 1);
    }
    sector_num++;
    break;

  case 'b':
    make_saucer("noname");
    set_saucers(0);
    break;

  case 'r':
  case 'e':
    n = ObjList_GetLength(saucers);
    n = rand() % n;

    for (obj = ObjList_GetStart(saucers);
	 !ObjList_IsEndEdge(saucers, obj);
	 obj = ObjListData_GetNext(obj)) {
      if (n == 0)
	break;
      n--;
    }
    saucer = (Saucer)ObjListData_GetObj(obj);

    t = rand() % Saucer_GetTrackNum(saucer);
    s = rand() % Saucer_GetSectorNum(saucer);

    block = Saucer_GetBlock(saucer, t, s);

    switch (key) {
    case 'r':
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_ATTACK);
      break;
    case 'e':
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_TRY);
      break;
    default:
      break;
    }
    break;

  case 's':
    set_saucers(1);
    break;

  case 'a':
    set_saucers(-1);
    break;

  default:
    break;
  }
}

static void gl_special_keyboard(int key, int x, int y)
{
  switch (key) {
  case GLUT_KEY_UP:    gl_keyboard('i', x, y); return;
  case GLUT_KEY_DOWN:  gl_keyboard('k', x, y); return;
  case GLUT_KEY_LEFT:  gl_keyboard('j', x, y); return;
  case GLUT_KEY_RIGHT: gl_keyboard('l', x, y); return;
  default: break;
  }
}

static void gl_reshape(int width, int height)
{
  glViewport(0, 0, width, height);

  glMatrixMode(GL_PROJECTION);

  glLoadIdentity();

  gluPerspective(30.0, (double)width / (double)height, 1.0, 100.0);

  glMatrixMode(GL_MODELVIEW);
}

static int socket_proc();

static void gl_timeout(int value)
{
  Saucer saucer;
  ObjListData obj;

  socket_proc();

  if (auto_rate > 0) {
    if ((rand() % 1000) < auto_rate) {
      if ((rand() % 10) == 0)
	gl_keyboard('r', 0, 0);
      else
	gl_keyboard('e', 0, 0);
    }
  }

  for (obj = ObjList_GetStart(saucers);
       !ObjList_IsEndEdge(saucers, obj);
       obj = ObjListData_GetNext(obj)) {
    saucer = (Saucer)ObjListData_GetObj(obj);
    Saucer_Proc(saucer);
  }

  camera.angle_dir += camera.angle_dir_v;
  glutPostRedisplay();
  glutTimerFunc(33, gl_timeout, 0);
}

static int accept_sockt = -1;
static ObjList sock_list = NULL;

static int ObjList_GetIndex(ObjList list, Obj object)
{
  ObjListData obj;
  int i = 0;
  if (list == NULL) return -1;
  for (obj = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, obj);
       obj = ObjListData_GetNext(obj)) {
    if (ObjListData_GetObj(obj) == object)
      return i;
    i++;
  }
  return -1;
}

static Obj ObjList_SearchIndex(ObjList list, int index)
{
  ObjListData obj;
  if (list == NULL) return NULL;
  for (obj = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, obj);
       obj = ObjListData_GetNext(obj)) {
    if (index == 0)
      return ObjListData_GetObj(obj);
    index--;
  }
  return NULL;
}

static int socket_proc()
{
  Saucer saucer;
  Block block;
  User user;
  Genre genre;
  Question question;
  ObjListData obj;
  int t, s;
  int sockt;
  char **argv;

  if (multi_sock || (ObjList_GetLength(sock_list) < 1)) {
    if (accept_sockt >= 0) {
      sockt = socket_accept_server(accept_sockt);
      if (sockt >= 0)
	SockList_AddSock(sock_list, sockt);
    }
  }

  argv = SockList_ReadLine(sock_list, 0);
  if ((argv == NULL) || (argv[0] == NULL))
    return -1;

  if (!strcmp(argv[0], "USER")) {
    if ((argv[1] == NULL) || (argv[2] == NULL))
      return -1;

    user = User_ListSearch(users, argv[1]);
    if (user) {
      saucer = User_GetSaucer(user);
      Saucer_SetName(saucer, argv[2]);
    } else {
      if (ObjList_GetLength(users) < ObjList_GetLength(saucers)) {
	saucer = (Saucer)ObjList_SearchIndex(saucers,
					     ObjList_GetLength(users));
      } else {
	saucer = make_saucer(NULL);
      }
      Saucer_SetName(saucer, argv[2]);
      Saucer_SetScoreImage(saucer);
      user = User_Create(argv[1], argv[2], saucer);
      ObjList_InsertObjToEnd(users, user, (ObjDestructor)User_Destroy);
      set_saucers(0);
    }
  }

  if (!strcmp(argv[0], "QUESTION")) {
    if ((argv[1] == NULL) || (argv[2] == NULL) || (argv[3] == NULL))
      return -1;

    genre = Genre_ListSearch(genres, argv[2]);
    if (genre == NULL) {
      genre = Genre_Create(argv[2]);
      ObjList_InsertObjToEnd(genres, genre, (ObjDestructor)Genre_Destroy);
      while (ObjList_GetLength(genres) > sector_num) {
	for (obj = ObjList_GetStart(saucers);
	     !ObjList_IsEndEdge(saucers, obj);
	     obj = ObjListData_GetNext(obj)) {
	  saucer = (Saucer)ObjListData_GetObj(obj);
	  Saucer_AddSectorNum(saucer, 1);
	}
	sector_num++;
      }
    }

    if (Question_ListSearch(Genre_GetQuestions(genre), argv[1]))
      return -1;

    question = Question_Create(argv[1], atoi(argv[3]), genre);
    ObjList_InsertObjToEnd(Genre_GetQuestions(genre), question,
			   (ObjDestructor)Question_Destroy);

    while (ObjList_GetLength(Genre_GetQuestions(genre)) > track_num) {
      for (obj = ObjList_GetStart(saucers);
	   !ObjList_IsEndEdge(saucers, obj);
	   obj = ObjListData_GetNext(obj)) {
	saucer = (Saucer)ObjListData_GetObj(obj);
	Saucer_AddTrackNum(saucer, 1);
      }
      track_num++;
    }

    s = ObjList_GetIndex(genres, (ObjListData)genre);
    t = ObjList_GetIndex(Genre_GetQuestions(genre), (ObjListData)question);
    if (!((s < 0) || (t < 0))) {
      for (obj = ObjList_GetStart(saucers);
	   !ObjList_IsEndEdge(saucers, obj);
	   obj = ObjListData_GetNext(obj)) {
	saucer = (Saucer)ObjListData_GetObj(obj);
	block = Saucer_GetBlock(saucer, t, s);
	Block_SetQuestion(block, question);
      }
    }
  }

  if (!strcmp(argv[0], "REPLY") || !strcmp(argv[0], "CLEAR")) {
    if ((argv[1] == NULL) || (argv[2] == NULL) || (argv[3] == NULL))
      return -1;

    user = User_ListSearch(users, argv[1]);
    if (user == NULL)
      return -1;
    question = Question_ListSearchAll(genres, argv[2]);
    if (question == NULL)
      return -1;
    genre = Question_GetGenre(question);
    saucer = User_GetSaucer(user);

    s = ObjList_GetIndex(genres, (ObjListData)genre);
    t = ObjList_GetIndex(Genre_GetQuestions(genre), (ObjListData)question);
    if ((s < 0) || (t < 0))
      return -1;

    block = Saucer_GetBlock(saucer, t, s);

    if (!strcmp(argv[0], "REPLY"))
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_TRY);
    if (!strcmp(argv[0], "CLEAR"))
      Block_AddCanon(block, Canon_Create(block), BLOCK_STATUS_ATTACK);
  }

  return 0;
}

static char *server_name = NULL;
static int port = 12345;
static int server_mode = 0;
static int udp_mode = 0;
static int stand_alone = 0;
static char *kanji_name = NULL;

static Argument args[] = {
  { "-host"      , ARGUMENT_TYPE_STRING  , &server_name },
  { "-h"         , ARGUMENT_TYPE_STRING  , &server_name },
  { "-port"      , ARGUMENT_TYPE_INTEGER , &port },
  { "-p"         , ARGUMENT_TYPE_INTEGER , &port },
  { "-server"    , ARGUMENT_TYPE_FLAG_ON , &server_mode },
  { "-s"         , ARGUMENT_TYPE_FLAG_ON , &server_mode },
  { "-udp"       , ARGUMENT_TYPE_FLAG_ON , &udp_mode },
  { "-u"         , ARGUMENT_TYPE_FLAG_ON , &udp_mode },
  { "-multi"     , ARGUMENT_TYPE_FLAG_ON , &multi_sock },
  { "-m"         , ARGUMENT_TYPE_FLAG_ON , &multi_sock },
  { "-alone"     , ARGUMENT_TYPE_FLAG_ON , &stand_alone },
  { "-saucer-num", ARGUMENT_TYPE_INTEGER , &saucer_num },
  { "-track-num" , ARGUMENT_TYPE_INTEGER , &track_num },
  { "-sector-num", ARGUMENT_TYPE_INTEGER , &sector_num },
  { "-auto-rate" , ARGUMENT_TYPE_INTEGER , &auto_rate },
  { "-kanji"     , ARGUMENT_TYPE_STRING  , &kanji_name },
  { NULL     , ARGUMENT_TYPE_NONE    , NULL },
};

int main(int argc, char *argv[])
{
  int sockt = -1;

  sock_list = ObjList_Create();

  argument_read(&argc, argv, args);

  if (!stand_alone) {
    if (server_mode) {
      if (!udp_mode)
	accept_sockt = socket_open_server(server_name, port);
      else
	sockt = socket_open_dgram_server(server_name, port);
    } else if (server_name) {
      sockt = socket_open_client(server_name, port);
    }
    if (sockt >= 0)
      SockList_AddSock(sock_list, sockt);
  }

  if (kanji_name) {
    kanji_init(kanji_name);
  }

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(800, 600);
  glutCreateWindow(argv[0]);

  glutDisplayFunc(gl_display);
  glutKeyboardFunc(gl_keyboard);
  glutSpecialFunc(gl_special_keyboard);
  glutReshapeFunc(gl_reshape);
  glutTimerFunc(33, gl_timeout, 0);

  init();

  glutMainLoop();
  return 0;
}
