#!/usr/bin/perl

# pcf2bdf /usr/local/lib/X11/fonts/local/jiskan16-2000-1.pcf.gz > kanji.bdf
# ./bdf2sfont.pl kanji.bdf > kanji.sfont
# ./saucer -kanji kanji.sfont

use Encode;

# UTF8: utf8
# EUC : euc-jp
# SJIS: shiftjis
# JIS : iso-2022-jp
$enc = "utf8";

$kanji_in  = "\x1B\x24\x42";
$kanji_out = "\x1B\x28\x42";

$width  = 16;
$height = 16;
$offset_x = 0;
$offset_y = 0;
$r = 0;

while (<>) {
	if (/^FONTBOUNDINGBOX/) {
		($width, $height, $offset_x, $offset_y) =
			/(\d+)\s+(\d+)\s+(\-?\d+)\s+(\-?\d+)/;
		next;
	}

	if (/^ENCODING/) {
		($jis_dec) = /(\d+)/;
		$jis_hex = sprintf("%04X", $jis_dec);
		$jis_char = $kanji_in . pack("H4", $jis_hex) . $kanji_out;
		$code_char = encode($enc, decode("iso-2022-jp", $jis_char));
		$code_hex = unpack("H*", $code_char);
		print "BEGIN $code_hex $width $height\n";
		next;
	}

	if (/^BBX/) {
		($w, $h, $x, $y) = /(\d+)\s+(\d+)\s+(\-?\d+)\s+(\-?\d+)/;
		$x = $x - $offset_x;
		$y = $height - ($h + ($y - $offset_y));
		next;
	}

	if (/^BITMAP/) {
		for ($j = 0; $j < $height; $j++) {
			for ($i = 0; $i < $width; $i++) {
				$bitmap[$j][$i] = 0;
			}
		}
		$r = 1;
		next;
	}

	if (/^ENDCHAR/) {
		if ($r) {
			for ($j = 0; $j < $height; $j++) {
				$h = "";
				$v = 0;
				$d = 0;
				$s = "";
				for ($i = 0; $i < $width; $i++) {
					if ($bitmap[$j][$i]) {
						$v |= (1 << (3 - $d));
						$s .= "*";
					} else {
						$s .= ".";
					}
					$d++;
					if ($d == 4) {
						$h .= sprintf("%X", $v);
						$v = 0;
						$d = 0;
					}
				}
				print "$h $s\n";
			}
			$r = 0;
		}
		print "END\n";
		next;
	}

	if ($r && /^[A-Fa-f0-9]/) {
		($bits) = /([A-Fa-f0-9]+)/;
		$d = $x;
		while (length($bits) > 0) {
			($b, $bits) = $bits =~ /^([A-Fa-f0-9])([A-Fa-f0-9]*)/;
			$b = hex($b);
			for ($i = 0; $i < 4; $i++) {
				if (($b >> (3 - $i)) & 1) {
					$bitmap[$y + $r - 1][$d + $i] = 1;
				}
			}
			$d += 4;
		}
		$r++;
		next;
	}
}
